unit ICCore;

(* Start Header *)

(* File:         ICCore.p
 * Generated by: 1.0d6
 * For:          IC 2.0
 * On:           Wednesday, 24 December 1997, 8:29:16
 * 
 * This file is part of the Internet Configuration system and
 * is placed in the public domain for the benefit of all.
 *)

(* End Header *)

(* This file contains the core implementation of the Internet Config API.
	It defines the format of the ICRRecord type (which the outside world
	sees indirectly by way of the ICInstance, a pointer to this record) and
	the various IC API operations on that type.
*)

interface

	uses
		Types,
		Files, 
		QuickDraw, 
		Aliases, 
		Memory,
		Components,
		
		ICSystemGlobals,
		InternetConfig;


	(* Copied Key Notes

		With version 1.4 and up, the ICAPI no longer requires that key
		strings be on word boundaries on 68000 machines.  This was 
		causing a problem for CodeWarrior C uses, because their
		compiler didn't align Pascal strings on word boundaries (duh!).
		Anyway, one of the tasks for 1.4 was to reduce the stack space
		usage, and fixing this key alignment problem just fell out
		of the wash.
		
		The external IC routines all use a Str255 interface, but
		the internal routines all use StringPtr instead.  This
		avoids the data copying.  When I need to use the string,
		I copy it into a local Str255 (typically called "copiedKey")
		using BlockMoveData.  BlockMoveData has no alignment 
		restrictions, so our 68000 machines are happy.
		
		In the process of doing this, I also guard myself against
		the key StringPtr being nil.  If any of my C programming
		friends try to do that, they'll get a nice error code.
	*)

	(* ***** Core API Routines ***** *)
	
	(*	These interfaces are automagically generated by the Internet Config
		Interface Generator.  See "InternetConfig.p" for a description of
		the behaviour of these routines.
	*)
	
(* Start ICCICore.p *)

	function ICCIStart(globals : ComponentGlobalsPtr; creator : OSType) : ICError;
	
	function ICCIStop(globals : ComponentGlobalsPtr) : ICError;
	
	function ICCIFindConfigFile(globals : ComponentGlobalsPtr; count : integer; folders : ICDirSpecArrayPtr) : ICError;
	
	function ICCIFindUserConfigFile(globals : ComponentGlobalsPtr; var where : ICDirSpec) : ICError;
	
	function ICCIGeneralFindConfigFile(globals : ComponentGlobalsPtr; searchPrefs : Boolean; canCreate : Boolean; count : integer; folders : ICDirSpecArrayPtr) : ICError;
	
	function ICCIChooseConfig(globals : ComponentGlobalsPtr) : ICError;
	
	function ICCIChooseNewConfig(globals : ComponentGlobalsPtr) : ICError;
	
	function ICCIGetConfigName(globals : ComponentGlobalsPtr; longname : Boolean; var name : Str255) : ICError;
	
	function ICCIGetConfigReference(globals : ComponentGlobalsPtr; ref : ICConfigRefHandle) : ICError;
	
	function ICCISetConfigReference(globals : ComponentGlobalsPtr; ref : ICConfigRefHandle; flags : longint) : ICError;
	
	function ICCISpecifyConfigFile(globals : ComponentGlobalsPtr; var config : FSSpec) : ICError;
	
	function ICCIRefreshCaches(globals : ComponentGlobalsPtr) : ICError;
	
	function ICCIGetSeed(globals : ComponentGlobalsPtr; var seed : longint) : ICError;
	
	function ICCIGetPerm(globals : ComponentGlobalsPtr; var perm : ICPerm) : ICError;
	
	function ICCIDefaultFileName(globals : ComponentGlobalsPtr; var name : Str63) : ICError;
	
	function ICCIBegin(globals : ComponentGlobalsPtr; perm : ICPerm) : ICError;
	
	function ICCIGetPref(globals : ComponentGlobalsPtr; key : StringPtr; var attr : ICAttr; buf : Ptr; var size : longint) : ICError;
	
	function ICCISetPref(globals : ComponentGlobalsPtr; key : StringPtr; attr : ICAttr; buf : Ptr; size : longint) : ICError;
	
	function ICCIFindPrefHandle(globals : ComponentGlobalsPtr; key : StringPtr; var attr : ICAttr; prefh : Handle) : ICError;
	
	function ICCIGetPrefHandle(globals : ComponentGlobalsPtr; key : StringPtr; var attr : ICAttr; var prefh : Handle) : ICError;
	
	function ICCISetPrefHandle(globals : ComponentGlobalsPtr; key : StringPtr; attr : ICAttr; prefh : Handle) : ICError;
	
	function ICCICountPref(globals : ComponentGlobalsPtr; var count : longint) : ICError;
	
	function ICCIGetIndPref(globals : ComponentGlobalsPtr; index : longint; var key : Str255) : ICError;
	
	function ICCIDeletePref(globals : ComponentGlobalsPtr; key : StringPtr) : ICError;
	
	function ICCIEnd(globals : ComponentGlobalsPtr) : ICError;
	

(* End ICCICore.p *)

	function CheckInside (globals : ComponentGlobalsPtr; requiredPerm: ICPerm): ICError;

	function ForceInside (globals : ComponentGlobalsPtr; perm: ICPerm; var forceInsideInfo: Boolean): ICError;
	function ReleaseInside (globals : ComponentGlobalsPtr; forceInsideInfo: Boolean): ICError;

	procedure DirtyThisFile(globals : ComponentGlobalsPtr);
	
implementation

	uses
		Resources,
		GestaltEqu, 
		OSUtils, 
		ToolUtils, 
		StandardFile, 
		AppleTalk, 
		Folders, 
		Processes, 
		Errors, 
		AppleEvents, 
		LowMem,

		ICDebug,
		ICStringLists,
		ICCommonSubs,
		ICResourceForkSanity, 
		ICComponentResources,
		ICProfiles,
		ICMapping;

	function CloseIfOpen (globals : ComponentGlobalsPtr): Boolean;
		forward;

	(* ***** Utility Routines ***** *)

	const
		kCopyRsrcToData = true;
		kCopyDataToRsrc = false;
		
	function CopyFileForkToFork (var fss: FSSpec; rsrcToData: Boolean): OSErr;
		(* This routine is a simple wrapper around CopyForkToFork.  It is
			used to make backups of the resource fork in the data fork,
			and to unbackup from the data fork to the resource fork.
			fss is the file to backup.  If rsrcToData is true, it copies the resource
			fork to the data fork, otherwise vice versa.
		*)
	begin
		CopyFileForkToFork := CopyForkToFork(fss, fss, rsrcToData, not rsrcToData);
	end; (* CopyFileForkToFork *)

	function BackupFileIntoTrash (globals: ComponentGlobalsPtr): OSErr;
		(* Make a copy of the preferences file in the Trash.  This function
			is called when IC detects that a preferences file is corrupt.
			Instead of just overwriting the file, it copies it to the Trash
			and then overwrites it.  This way, the user (or, more importantly, I)
			can analyse the file to see what symptoms it displays.
		*)
		var
			err: OSErr;
			junk: OSErr;
			createdBackupFile: Boolean;
			damagedFileCatalogInfo: CInfoPBRec;
			backupFile: FSSpec;
			tries: integer;
			middlePartString: Str31;
			triesString: Str255;
	begin
		(* Prepare for failure.  We remember whether we created a backup
			file so that we can delete it on failure at the end of this routine.
		*)
		createdBackupFile := false;

		err := noErr;

		(* Get info about the damaged file (ie the currently configured file) to
			use when we create the copy.
		*)
		if err = noErr then begin
			err := FSpGetCatInfo(globals^.configFileSpec, 0, damagedFileCatalogInfo);
		end; (* if *)

		(* Now create an FSSpec that points to the file we'd like to create in the Trash. *)
		if err = noErr then begin
			backupFile.name := globals^.configFileSpec.name;
			err := FindFolder(kOnSystemDisk, kTrashFolderType, kCreateFolder, backupFile.vRefNum, backupFile.parID);
		end; (* if *)

		(* Try and create the file in the Trash.  This involves creating it and, if we fail,
			adding an incrementing numeric index to the end of the name.  We do this
			up to 100 times before bailing out completely.
		*)
		if err = noErr then begin
			middlePartString := ' (Damaged)';
			junk := GetComponentIndString(globals, rMiscStrings, strDamagedSuffix, @middlePartString);

			tries := 0;
			triesString := '';
			repeat
				(* Do some magic to merge the various file names.  Basically we have
					3 components that we'd like to concatenate:
						1. the damaged file name
						2. the middle part, ie " (Damaged)"
						3. the tries string, ie "" the first time, " 1" the second time, and so on.
					We first try concating them to see if they'll all find into 31 characters.
					If not, we reduce the length of the first component (the damaged file
					name) until the name fits into 31 characters.
				*)
				if length(concat(globals^.configFileSpec.name,
											middlePartString, 
											triesString)) > 31 then begin
					backupFile.name := concat(copy(globals^.configFileSpec.name,
																		1, 
																		31 - length(middlePartString) - length(triesString)), 
																middlePartString, 
																triesString);
				end else begin
					backupFile.name := concat(globals^.configFileSpec.name, 
																middlePartString, 
																triesString);
				end; (* if *)
				err := FSpCreate(backupFile, damagedFileCatalogInfo.ioFlFndrInfo.fdCreator, damagedFileCatalogInfo.ioFlFndrInfo.fdType, smSystemScript);
				tries := tries + 1;
				NumToString(tries, triesString);
				triesString := concat(' ', triesString);
			until (err = noErr) or (tries > 100);

		end; (* if *)

		(* Copy the forks to the newly created file, then set it's Finder info
			appropriately.
		*)
		if err = noErr then begin
			createdBackupFile := true;
			err := CopyForkToFork(globals^.configFileSpec, backupFile, false, false);
		end; (* if *)
		if err = noErr then begin
			err := CopyForkToFork(globals^.configFileSpec, backupFile, true, true);
		end; (* if *)
		if err = noErr then begin
			(* Clear the "inited" bit. *)
			damagedFileCatalogInfo.ioFlFndrInfo.fdFlags := band(damagedFileCatalogInfo.ioFlFndrInfo.fdFlags, $FFFFFEFF);
			err := FSpSetCatInfo(backupFile, damagedFileCatalogInfo);
		end; (* if *)

		(* Clean up. *)
		if (err <> noErr) and createdBackupFile then begin
			junk := FSpDelete(backupFile);
		end; (* if *)

		BackupFileIntoTrash := err;
	end; (* BackupFileIntoTrash *)
	
	procedure DisplayFileFateDialog (globals: ComponentGlobalsPtr;
															irrecoverably: Boolean; 
															backupMade: Boolean);
		(* Display a dialog showing the user the fate of their Internet Preferences file.
			If irrecoverably is set, say that the file was reverted to defaults, otherwise
			say it was merely reverted to a backup.  If backupMade is set, say the damaged
			copy is in the Trash, otherwise say nothing.
		*)
		var
			err: OSErr;
			junk: integer;
			backupMadeMessage: Str255;
			damageReport : Str255;
			didOpen : Boolean;
	begin
		err := ICUCanInteract;
		if err = noErr then begin
			err := StartUsingComponentResources(globals, didOpen);
		end; (* if *)
		if err = noErr then begin
			if backupMade then begin
				junk := GetComponentIndString(globals, rMiscStrings, strBackupInTrash, @backupMadeMessage);
			end else begin
				backupMadeMessage := '';
			end; (* if *)
			damageReport := '';
			junk := GetComponentIndString(globals, rMiscStrings, strDamaged + ord(irrecoverably), @damageReport);
			ParamText(globals^.configFileSpec.name, damageReport, backupMadeMessage, '');
			junk := NoteAlert(rDamagedAlert, nil);
			
			junk := StopUsingComponentResources(globals, didOpen);
		end; (* if *)
	end; (* DisplayFileFateDialog *)

	(* ***** Starting Up ***** *)

	function ICCIStart (globals: ComponentGlobalsPtr; creator: OSType): ICError;
		(* Handle the start request.  This is basically a replacement for the Open
			because we need bonus data (creator).  This implementation calls though
			to the link-in implementation and then grabs some resources that
			we need to have conveniently at hand.
		*)
		var
			err: OSErr;
			didOpen : Boolean;
			tmpH: Handle;
			junk: OSErr;
	begin
		{$unused creator}
		
		(*	Initialise all the fields in the instance.  Some fields are
			initialised through resources.
		*)
		
		(* General flags. *)
		
		globals^.haveConfigFile := false;
		globals^.waitingForFirstBegin := false;
		globals^.thisFileModified := false;
		
		(* Config file info. *)
		
		globals^.configFileSpec.vRefNum := 0;
		globals^.configFileSpec.parID := 0;
		globals^.configFileSpec.name := '';
		globals^.configFileRefNum := 0;
		globals^.configFilePerm := icNoPerm;
		
		(* Component resources. *)
		
		globals^.oldResourceFileRefNum := 0;
		globals^.resourceFileRefNum := 0;

		// System-wide globals
		
		globals^.sysGlobals := FindICSystemGlobals(globals);
		if globals^.sysGlobals = nil then begin
		
			ComponentWatchPoint(globals, kICDebugSignificantEventsMask, 'System globals created at Start time', 0);
		
			err := RegisterSystemGlobals(globals);
			if err = noErr then begin
				globals^.sysGlobals := FindICSystemGlobals(globals);
				if globals^.sysGlobals = nil then begin
					ICDebugStr('Could not create system globals at start time');
					err := icInternalErr;
				end; (* if *)
			end; (* if *)
		end; (* if *)
		
		(* Stuff from resources. *)
				
		err := StartUsingComponentResources(globals, didOpen);
		if err = noErr then begin
				
			(* Setup our global knowledge about the component's version number. *)
			
			tmpH := Get1Resource('vers', 1);
			err := CheckResError(tmpH);
			if err = noErr then begin
				globals^.componentVersion := VersRecHndl(tmpH)^^.numericVersion;
			end; (* if *)

			(* Default file name. *)
			
			if err = noErr then begin			
				err := ICCDefaultFileName(globals^.target, globals^.defaultFileName);
			end; (* if *)

			(* Set up the icApplicationCreator from a resource. *)

			if err = noErr then begin
				tmpH := Get1Resource('DATA', rICApplicationCreator);
				err := CheckResError(tmpH);
				if err = noErr then begin
					globals^.icApplicationCreator := OSTypePtr(tmpH^)^;
				end; (* if *)
			end; (* if *)

			junk := StopUsingComponentResources(globals, didOpen);
		end; (* if *)
		globals^.haveStarted := (err = noErr);
		ICCIStart := err;
	end; (* ICCIStart *)

	(* ***** Shutting Down ***** *)

	function ICCIStop (globals: ComponentGlobalsPtr): ICError;
		(* The component version of the Stop routine detects whether we're
			stopping an instance that points to a modified file.  If we are,
			it backs up the resource fork in the file's data fork.
		*)
		var
			err : OSStatus;
			junk: OSStatus;
	begin
		ICAssert(globals^.sysGlobals <> nil);
		err := noErr;
		if globals^.thisFileModified then begin
			ComponentWatchPoint(globals, kICDebugSignificantEventsMask, 'Making data fork backup', 0);
			junk := CopyFileForkToFork(globals^.configFileSpec, kCopyRsrcToData);
			
			// The IC application clones the preferences file into the Temporary
			// Items folder and targets IC at that file.  When you quit without
			// saving, it targets the main file, then retargets the temporary
			// document to flush out any changes.  Then, because you're not saving,
			// it deletes the temporary document.  Then, as it quits, it calls
			// this routine, which tries to back up the document in the
			// Temporary Items folder, which the application has already deleted.
			//
			// The long term solution to this problem is to have ICAp not clone
			// the file, but instead hold changes in memory while the user
			// edits the file, and then write them to the real database when
			// they save.  However, this requires a bunch of modifications to
			// ICAp, something that I just don't have time for right now.
			// So, instead, I'm hacking this assert to not error in that case.
			//
			// Quinn, 24 Dec 1997, IC 2.0a4
			
			ICAssert((junk = noErr) | (junk = fnfErr));
		end; (* if *)
		if globals^.registeredAsInterruptSafe then begin
			globals^.sysGlobals^.interruptSafeCount := globals^.sysGlobals^.interruptSafeCount - 1;
			ICAssert(globals^.sysGlobals^.interruptSafeCount >= 0);
			if globals^.sysGlobals^.interruptSafeCount = 0 then begin

				ComponentWatchPoint(globals, kICDebugSignificantEventsMask, 'Disposing interrupt safe mappings', 0);
				
				// Ther are no interrupt safe clients left running, so this
				// dispose and nil pair does not have to be atomic.
				
				DisposePtr(globals^.sysGlobals^.mappings);
				ICAssert(MemError = noErr);

				globals^.sysGlobals^.mappings := nil;
				globals^.sysGlobals^.mappingsSize := 0;
			end; (* if *)
		end; (* if *)
		if CheckInside(globals, icReadOnlyPerm) = noErr then begin
			ComponentWatchPoint(globals, kICDebugSignificantEventsMask, 'Stop is forcing an End', 0);
			err := ICCIEnd(globals);
		end; (* if *)
		
		// We've done our best to stop, so we mark ourselves as stopped.  The alternative
		// implementation of globals^.haveStarted := (err = noErr) would simply cause
		// the ICCIClose routine to call us again, but we've already done our best effort
		// at stopping, so there's no point trying again.
		
		globals^.haveStarted := false;
		
		ICCIStop := err;
	end; (* ICCIStop *)

	(* ***** Component Specific Routines ***** *)
	
	(* These are the API routines that the IC component overrides, to add
		some value on top of the link-in implementation that most people
		use.
	*)

	function ICCIDefaultFileName (globals: ComponentGlobalsPtr; var name: Str63): ICError;
		(* Gets the default file name for Internet Preferences files.  The component
			version of this routine gets that name out of a resource.
		*)
		var
			err : OSStatus;
	begin
		err := GetComponentIndString(globals, rMiscStrings, strDefaultFileName, @name);
		ICCIDefaultFileName := err;
	end; (* ICCIDefaultFileName *)

	(* ***** Specifying a Configuration ***** *)

	function ValidDirSpec (folder: ICDirSpec): ICError;
		(* Returns noErr if the ICDirSpec passed in points to an existing directory.
			This routine is called to check that the ICDirSpecs passed into the various
			configuration routines are valid before we accept them.
		*)
		var
			cpb: CInfoPBRec;
	begin
		cpb.ioVRefNum := folder.vRefNum;
		cpb.ioDirID := folder.dirID;
		cpb.ioNamePtr := nil;
		cpb.ioFDirIndex := -1;
		ValidDirSpec := PBGetCatInfoSync(@cpb);
	end; (* ValidDirSpec *)

	function ScanFolder (globals : ComponentGlobalsPtr; folder: ICDirSpec; var fileWeFound: FSSpec): Boolean;
		(* Scans a folder specified by the ICDirSpec looking for a valid Internet Preferences
			file.  The algorithm is a) first search for a file called Internet Preferences,
			b) then scan the entire folder looking for a type/creator match, and c)
			each time you check a file, you should resolve it if it's an alias.
		*)

		function FoundFile (folder: ICDirSpec; fileIndex: integer; var fileWeFound: FSSpec): OSErr;
			(* Check the file pointed to by either:

				a) if fileIndex = 0, folder.vRefNum + folder.dirID + fileWeFound.name
				b) if fileIndex > 0, folder.vRefNum + folder.dirID + fileIndex

				If the routine succeeds it returns noErr and fileWeFound is set to
				the file we found.  If the routine fails, it either returns the real
				error code, or a synthetic error code of 1, which means keep 
				incrementing the index and call me again.
			*)
			var
				err: OSErr;
				cpb: CInfoPBRec;
				isFolder: Boolean;
				junkWasAlias: Boolean;
		begin
			cpb.ioVRefNum := folder.vRefNum;
			cpb.ioDirID := folder.dirID;
			cpb.ioNamePtr := @fileWeFound.name;
			cpb.ioFDirIndex := fileIndex;
			err := PBGetCatInfoSync(@cpb);
			if err = noErr then begin
				
				(* If we get no error, we have to decide whether this is a useful file
					or not.  If it is, we continue to return noErr and that's the file we
					return.  If it isn't, we return 1, which tells the client to keep
					incrementing the index.
				*)
				fileWeFound.vRefNum := cpb.ioVRefNum;
				fileWeFound.parID := cpb.ioFlParID;
				if (btst(cpb.ioFlAttrib, 4) or (cpb.ioFlFndrInfo.fdType <> ICfiletype)) then begin
					err := 1; 				(* Folder, or not of the right type. *)
				end else begin
					err := ResolveAliasFile(fileWeFound, true, isFolder, junkWasAlias);
					if (err <> noErr) or isFolder then begin
						err := 1;			(* Broken alias, or an alias to a folder. *)
					end; (* if *)
				end; (* if *)
			end; (* if *)
			FoundFile := err;
		end; (* FoundFile *)

		var
			err: ICError;
			found: Boolean;
			fileIndex: integer;
	begin
		(* First look for the Internet Preferences file in the folder. *)
		fileWeFound.name := globals^.defaultFileName;
		found := (FoundFile(folder, 0, fileWeFound) = noErr);
		
		(* If we didn't find it, scan through all the files in the folder looking
			for a matching file type.
		*)
		if not found then begin
			fileIndex := 1;
			repeat
				fileWeFound.name := '';
				err := FoundFile(folder, fileIndex, fileWeFound);
				fileIndex := fileIndex + 1;
			until err <> 1;
			found := (err = noErr);
		end; (* if *)
		ScanFolder := found;
	end; (* ScanFolder *)

	(* ***** Specifying a Configuration ***** *)

	function ICCIFindConfigFile (globals : ComponentGlobalsPtr; count: integer; folders: ICDirSpecArrayPtr): ICError;
		(* See comment in ICAPI.p. *)
	begin
		ICCIFindConfigFile := ICCGeneralFindConfigFile(globals^.target, 
										true, true, 
										count, folders);
	end; (* ICCIFindConfigFile *)

	function ICCIFindUserConfigFile (globals : ComponentGlobalsPtr; var where: ICDirSpec): ICError;
		(* See comment in ICAPI.p. *)
	begin
		ICCIFindUserConfigFile := ICCGeneralFindConfigFile(globals^.target, 
															false, true, 
															1, @where);
	end; (* ICCIFindUserConfigFile *)

	function ICCIGeneralFindConfigFile (globals : ComponentGlobalsPtr; searchPrefs: Boolean; canCreate: Boolean; count: integer; folders: ICDirSpecArrayPtr): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			dirSpecIndex: integer;
			found: Boolean;
			prefFolder: ICDirSpec;
			lastFolderScanned: ICDirSpec;
			tempConfigFileSpec: FSSpec;
	begin
		(* Check the parameters. *)
		err := noErr;
		if (err = noErr) & (globals^.configFilePerm <> icNoPerm) then begin
			err := paramErr;					(* Must not be inside Begin/End pair. *)
		end; (* if *)
		if (err = noErr) & ((count < 0) | ((count <> 0) & (folders = nil))) then begin
			err := paramErr;					(* Check that count and folders makes sense. *)
		end; (* if *)
		if (err = noErr) & (count = 0) & not searchPrefs & canCreate then begin
			err := paramErr;					(* Being told to create a file without specifying where. *)
		end; (* if *)
		(* Check that all the specified directories exist. *)
		dirSpecIndex := 0;
		while (err = noErr) & (dirSpecIndex < count) do begin
			err := ValidDirSpec(folders^[dirSpecIndex]);
			dirSpecIndex := dirSpecIndex + 1;
		end; (* for *)
		
		(* OK, parameters are cool, let's start the job. *)
		if err = noErr then begin
		
			(* Search all the specified directories. *)
			dirSpecIndex := 0;
			found := false;
			while (dirSpecIndex < count) and not found do begin
				found := ScanFolder(globals, folders^[dirSpecIndex], tempConfigFileSpec);
				lastFolderScanned := folders^[dirSpecIndex];
				dirSpecIndex := dirSpecIndex + 1;
			end; (* while *)
			
			(* Now, if we haven't found it yet and we're asked to,
				search the preferences folder.
			*)
			if not found & searchPrefs then begin
				err := FindFolder(kOnSystemDisk, kPreferencesFolderType, kCreateFolder,
										prefFolder.vRefNum, 
										prefFolder.dirID);
				if (err = noErr) then begin
					found := ScanFolder(globals, prefFolder, tempConfigFileSpec);
					lastFolderScanned := prefFolder;
				end;
			end; (* if *)
			
			(* Finally, if we still can't find it and we're allowed to, create the
				file in the last folder we scanned.
			*)
			if not found and canCreate then begin
				tempConfigFileSpec.vRefNum := lastFolderScanned.vRefNum;
				tempConfigFileSpec.parID := lastFolderScanned.dirID;
				tempConfigFileSpec.name := globals^.defaultFileName;
				found := true;
			end; (* if *)
			
			(* If we still can't find it, return an error. *)
			if not found then begin
				err := icConfigNotFoundErr;
			end; (* if *)
		end; (* if *)
		
		(* Now pass what we found along to ICSpecifyConfigFile. *)
		if err = noErr then begin
			err := ICCSpecifyConfigFile(globals^.target, tempConfigFileSpec);
		end; (* if *)
		ICCIGeneralFindConfigFile := err;
	end; (* ICCIGeneralFindConfigFile *)

	(* ***** Standard File Utilities ***** *)

	procedure SetSFCWD (globals : ComponentGlobalsPtr);
		(* Point Standard File at the directory that contains the instance's current
			config file, if there is one.
			
			Note that this most probbaly fails with newer (post 7.5) versions of
			Standard File, but I'm not sure whether it's worth doing the workaround
			at this time.
		*)
	begin
		if globals^.haveConfigFile then begin
			LMSetCurDirStore( globals^.configFileSpec.parID );
			LMSetSFSaveDisk( -globals^.configFileSpec.vRefNum );
		end; (* if *)
	end; (* SetSFCWD *)

	function GetFile (globals : ComponentGlobalsPtr; var chosenFileSpec: FSSpec): ICError;
		(* Prompt the user to choose a file using the System 6 or 7
			SF dialogs.
		*)
		var
			err: ICError;
			typeList: SFTypeList;
			newReply: StandardFileReply;
	begin
		SetSFCWD(globals);
		err := userCanceledErr;
		typeList[0] := ICfiletype;
		StandardGetFile(nil, 1, @typeList, newReply);
		if newReply.sfGood then begin
			chosenFileSpec := newReply.sfFile;
			err := noErr;
		end; (* if *)
		GetFile := err;
	end; (* GetFile *)

	function PutFile (globals : ComponentGlobalsPtr; var chosenFileSpec: FSSpec): ICError;
		(* Prompt the user to create a new file using the System 6 or 7
			SF dialogs.
		*)
		var
			err: ICError;
			typeList: SFTypeList;
			newReply: StandardFileReply;
			defaultFileName: Str63;
			promptString : Str255;
	begin
		SetSFCWD(globals);
		typeList[0] := ICfiletype;
		
		(* Setup the default file name, either with the name of the current file 
			or the default file name.
		*)
		if globals^.haveConfigFile then begin
			defaultFileName := globals^.configFileSpec.name;
		end else begin
			defaultFileName := globals^.defaultFileName;
		end; (* if *)
		
		(* Get the prompt from a component resource. *)
		
		err := GetComponentIndString(globals,  rMiscStrings, strNewConfigPrompt, @promptString);

		(* Ask the user where to save the file. *)
		
		if err = noErr then begin
			StandardPutFile(promptString, defaultFileName, newReply);
			if newReply.sfGood then begin
				chosenFileSpec := newReply.sfFile;
				err := noErr;
			end else begin
				err := userCanceledErr;
			end; (* if *)
		end; (* if *)
		PutFile := err;
	end; (* PutFile *)

	(* ***** Choosing a Configuration ***** *)

	function ICCIChooseConfig (globals : ComponentGlobalsPtr): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: OSErr;
			chosenFileSpec: FSSpec;
	begin
		(* Check parameters. *)
		err := noErr;
		if (err = noErr) & (globals^.configFilePerm <> icNoPerm) then begin
			err := paramErr;					(* Must not be inside Begin/End pair. *)
		end; (* if *)
		
		(* Check whether we can pose dialogs, ask the user to specify a file,
			then specify that config file.
		*)
		if err = noErr then begin
			err := ICUCanInteract;
		end; (* if *)
		if err = noErr then begin
			err := GetFile(globals, chosenFileSpec);
		end; (* if *)
		if err = noErr then begin
			err := ICCSpecifyConfigFile(globals^.target, chosenFileSpec);
		end; (* if *)
		ICCIChooseConfig := err;
	end; (* ICCIChooseConfig *)

	function ICCIChooseNewConfig (globals : ComponentGlobalsPtr): ICError;
		(* See comment in ICAPI.p. *)
		var
			err, junk: OSErr;
			chosenFileSpec: FSSpec;
	begin
		(* Check parameters. *)
		err := noErr;
		if (err = noErr) & (globals^.configFilePerm <> icNoPerm) then begin
			err := paramErr;					(* Must not be inside Begin/End pair. *)
		end;

		(* Check whether we can pose dialogs, ask the user to specify a new file,
			delete and recreate the file, then specify that config file.
		*)
		if err = noErr then begin
			err := ICUCanInteract;
		end; (* if *)
		if err = noErr then begin
			err := PutFile(globals, chosenFileSpec);
		end; (* if *)
		if err = noErr then begin
			junk := FSpDelete(chosenFileSpec);
			err := FSpCreate(chosenFileSpec, ICcreator, ICfiletype, smSystemScript);
		end; (* if *)
		if err = noErr then begin
			err := ICCSpecifyConfigFile(globals^.target, chosenFileSpec);
		end; (* if *)
		ICCIChooseNewConfig := err;
	end; (* ICCIChooseNewConfig *)

	(* ***** Other Configuration API Routines ***** *)

	function ICCIGetConfigName (globals : ComponentGlobalsPtr; longname: Boolean; var name: Str255): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: OSErr;
	begin
		(* Check parameters. *)
		err := noErr;
		if not globals^.haveConfigFile then begin
			err := paramErr;					(* Must have a configuration. *)
		end; (* if *)

		if longname then begin
			err := ICUFSSPecToFullPath(globals^.configFileSpec, name);
		end else begin
			name := globals^.configFileSpec.name;
			err := noErr;
		end; (* if *)
		ICCIGetConfigName := err;
	end; (* ICCIGetConfigName *)

	function ICCIGetConfigReference (globals : ComponentGlobalsPtr; ref: ICConfigRefHandle): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			configRefHeader: ICConfigRef;
			junkLong: longint;
	begin
		err := noErr;
		if not globals^.haveConfigFile then begin
			err := paramErr;					(* Must have a configuration. *)
		end; (* if *)
		if (err = noErr) & (ref = nil) then begin
			err := paramErr;					(* Client must supply handle for results. *)
		end; (* if *)
		
		(* First create an ICFileSpec to the file.  This contains both a real alias
			and a poor man's alias for System 6 operation.
		*)
		if err = noErr then begin
			err := FSSpecToICFileSpec(globals^.configFileSpec, ICFileSpecHandle(ref));
		end; (* if *)
		
		(* Now insert our manufacturer code at the front of the ICFileSpec
			to make it a config reference.
		*)
		if err = noErr then begin
			configRefHeader.manufacturer := kICComponentManufacturer;
			junkLong := Munger(Handle(ref), 0, nil, 0, @configRefHeader, sizeof(configRefHeader));
			err := MemError;
		end; (* if *)
		
		(* Clean up. *)
		if err <> noErr then begin
			SetHandleSize(Handle(ref), 0);
		end; (* if *)
		ICCIGetConfigReference := err;
	end; (* ICCIGetConfigReference *)

	function ICCISetConfigReference (globals : ComponentGlobalsPtr; ref: ICConfigRefHandle; flags: longint): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			filespec: ICFileSpecHandle;
			junkLong: longint;
			newConfigFileSpec: FSSpec;
	begin
		(* Check parameters. *)
		err := noErr;
		if globals^.configFilePerm <> icNoPerm then begin
			err := paramErr;					(* Must not be inside Begin/End pair. *)
		end; (* if *)
		if (err = noErr) & (ref = nil) then begin
			err := paramErr;					(* Client must supply handle for results. *)
		end; (* if *)
		if (err = noErr) & (GetHandleSize(Handle(ref)) < 4) then begin
			err := paramErr;					(* Handle must be a reasonable size. *)
		end; (* if *)
		if (err = noErr) & (ref^^.manufacturer <> kICComponentManufacturer) then begin
			err := icConfigInappropriateErr;	(* Config ref header must indicate that the config ref is ours. *)
		end; (* if *)
		if (err = noErr) & (GetHandleSize(Handle(ref)) < sizeof(ICConfigRef) + sizeof(ICFileSpec)) then begin
			err := paramErr;					(* Config ref data must be big enough for minimal ICFileSpec *)
		end; (* if *)

		(* Now do the work. *)
		if err = noErr then begin
			(* Make a copy of the handle. *)
			filespec := ICFileSpecHandle(ref);
			err := HandToHand(Handle(filespec));
			if err = noErr then begin
			
				(* Delete the config ref header off the front of the handle. *)
				junkLong := Munger(Handle(filespec), 0, nil, sizeof(ICConfigRef), @junkLong, 0);

				(* Pass the remaining handle to our 'resolve alias' code. *)
				err := ICFileSpecToFSSpec(filespec, not btst(flags, icNoUserInteraction_bit), newConfigFileSpec);
				if err = fnfErr then begin
					(* Swallow fnfErr because it might be a reference to a file that
						had never been created in the first place.
					*)
					err := noErr;
				end; (* if *)
				DisposeHandle(Handle(filespec));
			end; (* if *)
		end; (* if *)

		(* Use the new config file. *)
		if err = noErr then begin
			err := ICCSpecifyConfigFile(globals^.target, newConfigFileSpec);
		end; (* if *)
		ICCISetConfigReference := err;
	end; (* ICCISetConfigReference *)

	function ICCISpecifyConfigFile (globals: ComponentGlobalsPtr; var config: FSSpec): ICError;
		(* The component version of the SpecifyConfigFile routine detects
			whether we're pointing the instance away from a modified file.
			If we are, it backs up the resource fork in the file's data fork.
		*)
		var
			err : OSStatus;
			junk: OSStatus;
			folder : ICDirSpec;
	begin
		ICAssert(globals^.sysGlobals <> nil);
		
		(* Check parameters. *)
		err := noErr;
		if globals^.configFilePerm <> icNoPerm then begin
			err := paramErr;	(* Must not be inside Begin/End pair. *)
		end; (* if *)
		if globals^.registeredAsInterruptSafe then begin

			// You're not allowed to reconfigure an instance that
			// has been registered as an interrupt safe client.
			
			err := paramErr;
		end; (* if *)

		if globals^.thisFileModified then begin
			ComponentWatchPoint(globals, kICDebugSignificantEventsMask, 'Making data fork backup', 0);
			junk := CopyFileForkToFork(globals^.configFileSpec, kCopyRsrcToData);
		end; (* if *)
		
		(* Do the job.  Check that the new file is potentially valid (ie the file
			may not exist, but the directory it's in must and then switch
			configurations.
		*)
		if err = noErr then begin
			folder.vRefNum := config.vRefNum;
			folder.dirID := config.parID;
			err := ValidDirSpec(folder);
			if err = noErr then begin
				globals^.configFileSpec := config;
				globals^.waitingForFirstBegin := true;
				globals^.thisFileModified := false;
				globals^.haveDefaultConfig := EqualFSSpec(globals^.configFileSpec, 
												globals^.sysGlobals^.primaryConfigFile);
			end; (* if *)
			globals^.haveConfigFile := (err = noErr);
		end; (* if *)
		ICCISpecifyConfigFile := err;
	end; (* ICCISpecifyConfigFile *)
	
	(* ***** ***** *)
	
	function ICCIRefreshCaches(globals : ComponentGlobalsPtr) : ICError;
		var
			err : ICError;
	begin
		// Check parameters.
		
		err := noErr;
		if not globals^.haveConfigFile then begin
			err := paramErr;
		end; (* if *)
		
		// If we're a default instance, call RebuildInterruptSafeMapping
		// to refresh the interrupt safe mapping table.  Also bump the
		// primarySeed, so that clients know about the change through the
		// seed.  We don't just call DirtyThisFile because we don't need
		// to set thisFileModified because it will later force us to
		// make a backup of the file in the data fork, and the ICAp has
		// already done this before it copied the file over ours.
		
		if err = noErr then begin
			if globals^.haveDefaultConfig then begin
				globals^.sysGlobals^.primarySeed := globals^.sysGlobals^.primarySeed + 1;
				err := RebuildInterruptSafeMapping(globals);
			end; (* if *)
		end; (* if *)
		ICCIRefreshCaches := err;
	end; (* ICCIRefreshCaches *)

	(* ***** Getting Information ***** *)

	function ICCIGetSeed (globals : ComponentGlobalsPtr; var seed: longint): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			cpb: CInfoPBRec;
	begin
		seed := 0;
		if globals^.haveConfigFile then begin
			if globals^.haveDefaultConfig then begin
				ICAssert(globals^.sysGlobals <> nil);
				seed := globals^.sysGlobals^.primarySeed;
				err := noErr;
			end else begin
				cpb.ioVRefNum := globals^.configFileSpec.vRefNum;
				cpb.ioDirID := globals^.configFileSpec.parID;
				cpb.ioNamePtr := @globals^.configFileSpec.name;
				cpb.ioFDirIndex := 0;
				err := PBGetCatInfoSync(@cpb);
				if err = noErr then begin
					seed := cpb.ioFlMdDat;
				end else if err = fnfErr then begin
					err := noErr;
				end; (* if *)
			end; (* if *)
		end else begin
			err := fnfErr;
		end; (* if *)
		ICCIGetSeed := err;
	end; (* ICCIGetSeed *)

	function ICCIGetPerm (globals : ComponentGlobalsPtr; var perm: ICPerm): ICError;
		(* See comment in ICAPI.p. *)
	begin
		perm := globals^.configFilePerm;
		ICCIGetPerm := noErr;
	end; (* ICCIGetPerm *)

  (* ***** Begin/End Utilities ***** *)

	function PermToFSPerm (perm: ICPerm): integer;
		(* Maps an ICPerm to an file system permission. *)
	begin
		case perm of
			icReadOnlyPerm: 
				PermToFSPerm := fsRdPerm;
			icReadWritePerm: 
				PermToFSPerm := fsRdWrPerm;
			otherwise
				PermToFSPerm := 0;
		end; (* case *)
	end; (* PermToFSPerm *)

	function CheckInside (globals : ComponentGlobalsPtr; requiredPerm: ICPerm): ICError;
		(* The routine returns an error code based on whether you're
			inside a Begin/End pair. CheckInside is a utility used by API
			routines that require you to be inside a Begin/End pair to operate.
		*)
		var
			result : ICError;
	begin
		ICAssert(requiredPerm <> icNoPerm);
		result := noErr;
		case globals^.configFilePerm of
			icNoPerm:
				result := paramErr;
			icReadOnlyPerm:
				if requiredPerm = icReadWritePerm then begin
					result := paramErr;
				end; (* if *)
			icReadWritePerm:
				(* do nothing *) ;
		end; (* case *)
		CheckInside := result;
	end; (* CheckInside *)

	function ForceInside (globals : ComponentGlobalsPtr; requiredPerm: ICPerm; var forceInsideInfo: Boolean): ICError;
		(* This routine makes sure we're currently inside a Begin/End pair,
			calling Begin if necessary.  The routine sets forceInsideInfo to
			true if it called Begin, so that ReleaseInside knows whether
			to call End.  ForceInside is a utility used by API routines
			that automagically call Begin for you, such as GetPref.
		*)
		var
			err: ICError;
	begin
		ICAssert(requiredPerm <> icNoPerm);
		forceInsideInfo := false;
		if (globals^.configFilePerm = requiredPerm) or			(* We already have the required permission OR *)
				 ((globals^.configFilePerm = icReadWritePerm) 		(* We asking for read AND *)
				 			and (requiredPerm = icReadOnlyPerm))							(* We have read/write. *)
				 					then begin
			(* The permissions already match up, do nothing. *)
			err := noErr;
		end else if globals^.configFilePerm = icNoPerm then begin
			(* We have no existing permissions, call Begin and record whether
				we need to call End.
			*)
			err := ICCBegin(globals^.target, requiredPerm);
			forceInsideInfo := (err = noErr);
		end else begin
			(* If neither of the above cases work, fail with an error.  The only
				case that springs to mind is that we have read-only permissions
				and someone is asking for read/write.
			*)
			err := icPermErr;
		end; (* if *)
		ForceInside := err;
	end; (* ForceInside *)

	function ReleaseInside (globals : ComponentGlobalsPtr; forceInsideInfo: Boolean): ICError;
		(* Undoes the effect of a ForceInside. Note that unlike Begin/End, you
			should can safely call this regardless of whether ForceInside returns
			an error.
		*)
	begin
		if forceInsideInfo then begin
			ReleaseInside := ICCEnd(globals^.target);
		end else begin
			ReleaseInside := noErr;
		end; (* if *)
	end; (* ReleaseInside *)

	procedure DirtyThisFile(globals : ComponentGlobalsPtr);
	begin
		ICAssert(globals^.haveConfigFile);
		globals^.thisFileModified := true;
		if globals^.haveDefaultConfig then begin
			ICAssert(globals^.sysGlobals <> nil);
			globals^.sysGlobals^.primarySeed := globals^.sysGlobals^.primarySeed + 1;
		end; (* if *)
	end; (* DirtyThisFile *)

	function CloseIfOpen (globals : ComponentGlobalsPtr): Boolean;
		(* Close the instance's config file if it's open.  Returns
			true if the instance was in the right state for
			this routine to be called.
		*)
	begin
		CloseIfOpen := (globals^.configFilePerm <> icNoPerm);
		if globals^.configFileRefNum <> 0 then begin
			CloseResFile(globals^.configFileRefNum);
			globals^.configFileRefNum := 0;
		end; (* if *)
		globals^.configFilePerm := icNoPerm;
	end; (* CloseIfOpen *)

	(* ***** Begin/End ***** *)
	
	function CheckPreferenceFileSanity(globals : ComponentGlobalsPtr) : ICError;
		(* The component version of the Begin routine has special logic for
			detecting corrupt preferences file.  The logic goes:
			
				If it's the first time around, then invoke our special checks
					If the file is missing, we should create it
					if the file currently exists, we should validate it
						If the validation fails, we should
							Copy the file to the Trash
							Revert to backup
						If the reverted file still won't validate, we delete the entire thing
						Either way, we need to tell the user of the file's fate
					To finish up, we should AddComponentDefaultPrefs.
		*)
		var
			err : ICError;
			junk : ICError;
			fileFate : (ffNone, ffDamagedAndReverted, ffDamagedBeyondRepair);
			cpb : CInfoPBRec;
			sane : Boolean;
			refNum : integer;
			backupMade : Boolean;
	begin
		fileFate := ffNone;

		(* Check whether the file exists and has a resource fork.  If it fails,
			go create the file.
		*)
		err := FSpGetCatInfo(globals^.configFileSpec, 0, cpb);
		if (err = fnfErr) | ((err = noErr) & (cpb.ioFlRLgLen = 0)) then begin
			junk := FSpCreate(globals^.configFileSpec, ICcreator, ICfiletype, smSystemScript);
			FSpCreateResFile(globals^.configFileSpec, ICcreator, ICfiletype, smSystemScript);
		end; (* if *)

		(* Check the file for sanity.  If it fails, try to back it up
			in the Trash and then revert it to the the backup in the data fork.
		*)
		err := noErr;
		sane := false;
		junk := CheckResourceForkSanity(globals^.configFileSpec, sane);
		if not sane then begin
			backupMade := (BackupFileIntoTrash(globals) = noErr);
			fileFate := ffDamagedAndReverted;
			ComponentWatchPoint(globals, kICDebugSignificantEventsMask, 'Recovering from data fork backup', 0);
			err := CopyFileForkToFork(globals^.configFileSpec, kCopyDataToRsrc);
			if err = noErr then begin
				junk := CheckResourceForkSanity(globals^.configFileSpec, sane);
			end; (* if *)
		end; (* if *)

		(* Check the file for sanity again.  If it still fails, zap the file completely. *)
		if (err = noErr) & not sane then begin
			fileFate := ffDamagedBeyondRepair;
			err := FSpOpenRF(globals^.configFileSpec, fsRdWrPerm, refNum);
			if err = noErr then begin
				err := SetEOF(refNum, 0);
				junk := FSClose(refNum);
			end; (* if *)
			if err = noErr then begin
				junk := FSpCreate(globals^.configFileSpec, ICcreator, ICfiletype, smSystemScript);
				FSpCreateResFile(globals^.configFileSpec, ICcreator, ICfiletype, smSystemScript);
				err := ResError;
			end; (* if *)

			(* Check the file for sanity one more time, just to make sure we're
				not completely in the weeds.
			*)
			if err = noErr then begin
				DirtyThisFile(globals);
				junk := CheckResourceForkSanity(globals^.configFileSpec, sane);
				if not sane then begin
					err := icInternalErr;
				end; (* if *)
			end; (* if *)
		end; (* if *)

		(* Finally, let the user know what we did to their preferences, and
			make sure the file has the default preferences in it.
		*)
		if err = noErr then begin
			if fileFate <> ffNone then begin
				DisplayFileFateDialog(globals, fileFate = ffDamagedBeyondRepair, backupMade);
			end; (* if *)
		end; (* if *)
		CheckPreferenceFileSanity := err;
	end; (* CheckPreferenceFileSanity *)
	
	function AddDefaultPreferencesToCurrentProfile(globals : ComponentGlobalsPtr; 
									clientRequestedPerm : ICPerm;
									var resourceRefNum : integer) : ICError;
		(*	Adds the default preferences to the current profile (as set by the 'CURP' resource)
			in the configured preferences file.  resourceRefNum is only an output
			parameter.  If we decide to open the resource file, and our opening
			is compatible with clientRequestedPerm, then we leave it open
			and set resourceRefNum to the refnum we opened.
		*)
		var
			err : ICError;
			err2 : ICError;
			currentProfile : ICProfileID;
			currentProfileResource : Handle;
			prevTopMap : Handle;
			skipOperation : Boolean;
			{$ifc icDebug}
				writeable : Boolean;
			{$endc}
	begin
		err := noErr;
		skipOperation := false;
		
		// Attempt to open up the resource file.
		
		prevTopMap := LMGetTopMapHndl;
		resourceRefNum := FSpOpenResFile(globals^.configFileSpec, fsRdWrPerm);
		err := ResError;
		
		// In some circumstances (the resource file already being open in
		// resource chain of the current process), OpenResFile will not
		// open a new resource map, but will return the resourceRefNum set
		// to the existing open resource file.
		//
		// This only happens if you request write permissions.  The rest
		// of IC ignores this situation because it only requests the permissions
		// that were requested by the client, and the API *defines* that the
		// client shall not try to open a writer while a reader is open.
		// [IC should detect that case, but it's tricky, so we don't.  Still,
		// if the case arises, it's not our fault.]
		//
		// However, in this case, we're requesting read/write permission even if
		// the client only requested read-only permission, so we have to pay
		// attention to this potential problem.  The way to detect that Resource
		// Manager has not created a new resource refnum is to check for
		// a change in TopMapHndl (a low memory global that points to the top
		// resource map in the resource chain).
		//
		// If we open the file and TopMapHndl does not change, we know we've
		// hit this situation.  Our response is easy: we just skip the entire
		// AddDefaultPreferencesToCurrentProfile operation.  This works (in theory)
		// because the person who did the first Begin on some other instance that
		// points to this file will have already done the operation anyway.
		
		if prevTopMap = LMGetTopMapHndl then begin

			ComponentWatchPoint(globals, kICDebugSignificantEventsMask, 'OpenResFile returned a clone', 0);

			skipOperation := true;
			resourceRefNum := 0;
			err := noErr;
		end; (* if *)
		
		if (err = noErr) & not skipOperation then begin
			
			ICAssert((FileRefNumIsWriteable(CurResFile, writeable) = noErr) & writeable);
			
			// Now get the current profile resource, and use it to determine the
			// profile on which we should call AddDefaultPreferencesToProfile.
			// If there is no 'CURP' resource, we are dealing with an old style
			// (pre-2.0) preferences file, and we correct the 'CURP' resource by
			// setting it to the default.

			currentProfileResource := Get1Resource('CURP', rCurrentProfileResource);
			if (currentProfileResource = nil) | 
					(GetHandleSize(currentProfileResource) <> sizeof(ICProfileID)) |
					not IsICxxResource(ResTypePtr(currentProfileResource^)^) then begin
				
				// 'CURP' resource is missing or broken.  Recreate it.
				
				ComponentWatchPoint(globals, kICDebugSignificantEventsMask, 'Creating CURP resource for profile', 0);
				
				currentProfile := ICProfileID(kICPrefResourceTypeBase);
				err := Set1ResourcePtr(@currentProfile, sizeof(ICProfileID), 'CURP', rCurrentProfileResource);
				
				if err = noErr then begin
					DirtyThisFile(globals);
				end; (* if *)
				
			end else begin
			
				// 'CURP' resource seems to be OK.  Use it.
				
				currentProfile := ICProfileIDPtr(currentProfileResource^)^;
				
			end; (* if *)
			
			if err = noErr then begin
				err := AddDefaultPreferencesToProfile(globals, currentProfile);
			end; (* if *)
	
		end; (* if -- not skipOperation *)
		
		// Clean up.

		// If we successfully opened the resource file (we always open it
		// read/write) but the client only request read-only, we must close
		// the resource file so that the caller will open it again with
		// the right permissions.
		
		if (resourceRefNum <> 0) & (clientRequestedPerm <> icReadWritePerm) then begin
			CloseResFile(resourceRefNum);
			err2 := ResError;
			if err = noErr then begin
				err := err2;
			end; (* if *)
			resourceRefNum := 0;
		end; (* if *)
		
		AddDefaultPreferencesToCurrentProfile := err;
	end; (* AddDefaultPreferencesToCurrentProfile *)

	function ICCIBegin(globals : ComponentGlobalsPtr; perm: ICPerm): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			resourceRefNum: integer;
			currentProfileResource : Handle;
	begin
		resourceRefNum := 0;
		
		// Check parameters.
		
		err := noErr;
		if (globals^.configFilePerm <> icNoPerm) or (perm = icNoPerm) then begin
			err := paramErr;		(* Can't call Begin twice, or call it to ask for no permissions. *)
		end; (* if *)
		if err = noErr then begin
			if not globals^.haveConfigFile then begin
				err := bdNamErr;	(* Need to know where the file is do anything. *)
			end; (* if *)
		end; (* if *)

		// If this is the first Begin, check the resource file for sanity and then
		// check that it has the default preferences.  This will probably end up
		// opening the resource file, so we pass in resourceRefNum.  If this routine
		// sets resourceRefNum to non-zero, then it's already opened the resource
		// file, and we don't need to do it again later on in this code.
		
		if (err = noErr) & globals^.waitingForFirstBegin then begin
			err := CheckPreferenceFileSanity(globals);
			if err = noErr then begin
				err := AddDefaultPreferencesToCurrentProfile(globals, perm, resourceRefNum);
			end; (* if *)
		end; (* if *)
		
		// Open the resource file.

		if err = noErr then begin
			if resourceRefNum = 0 then begin
				resourceRefNum := FSpOpenResFile(globals^.configFileSpec, PermToFSPerm(perm));
				err := ResError;
			end; (* if *)
		end; (* if *)
		
		// If we've opened a resource file, then get the current profile
		// out of the 'CURP' resource in in the file.

		if err = noErr then begin
			currentProfileResource := Get1Resource('CURP', rCurrentProfileResource);
			err := CheckResError(currentProfileResource);
			if err = noErr then begin
				
				// All's well.  Let's narf the current profile info.
				
				globals^.configProfileResType := ResTypePtr(currentProfileResource^)^;
				ICAssert(IsICxxResource(globals^.configProfileResType));
				
			end else begin

				ICDebugStr('Should not have been able to get here.');
				
				// We should not be able to get here because the first Begin should
				// always have created the 'CURP' resource in AddDefaultPreferencesToCurrentProfile.
				// If we do get an error, we fail out to the caller.
				// But before we die, make sure we close the resource file to maintain
				// the behaviour of ICBegin (namely that it either opens the file
				// and succeeds, or it fails and leaves the file closed).
				
				CloseResFile(resourceRefNum);
			end; (* if *)
		end; (* if *)		

		// All is cool, record the fact that we've opened the file in the instance.
		
		if err = noErr then begin
			globals^.configFileRefNum := resourceRefNum;
			globals^.configFilePerm := perm;
			globals^.waitingForFirstBegin := false;
		end; (* if *)
		
		// Clean up the error result.
		
		case err of
			opWrErr, permErr: 
				err := icNoMoreWritersErr;
			otherwise
				(* do nothing *)
		end; (* case *)
		
		ICCIBegin := err;
	end; (* ICCIBegin *)

	function ICCIEnd (globals : ComponentGlobalsPtr): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			junkBool: Boolean;
	begin
		if globals^.profileCache <> nil then begin
			DisposeHandle(Handle(globals^.profileCache));
			ICAssert(MemError = noErr);
			globals^.profileCache := nil;
		end; (* if *)
		err := CheckInside(globals, icReadOnlyPerm);
		junkBool := CloseIfOpen(globals);
		ICCIEnd := err;
	end; (* ICCIEnd *)

	(* ***** Basic Reading and Writing Preferences ***** *)

	function ICCIGetPref (globals : ComponentGlobalsPtr; key: StringPtr; var attr: ICAttr; buf: Ptr; var size: longint): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			err2: ICError;
			copiedKey: Str255;
			maxSizeOfClientBuffer: longint;
			trueSizeOfPrefResource: longint;
			currentRefNum: integer;
			prefResourceH: Handle;
			forceInsideInfo: Boolean;
	begin
		(* Initialise our locals and return results to nice default values. *)
		copiedKey := '';
		maxSizeOfClientBuffer := size;
		size := 0;
		attr := ICattr_no_change;
		prefResourceH := nil;

		(* Make sure we're inside a read capable Begin/End pair. *)
		err := ForceInside(globals, icReadOnlyPerm, forceInsideInfo);

		(* Check other parameters. *)
		if (err = noErr) and (key <> nil) then begin
			BlockMoveData(key, @copiedKey, ord(key^[0]) + 1);
		end; (* if *)
		if (err = noErr) and ((copiedKey = '') or ((maxSizeOfClientBuffer < 0) and (buf <> nil))) then begin
			(* key was nil, or key was empty, or maxSizeOfClientBuffer was bogus without buf being nil. *)
			err := paramErr;
		end; (* if *)

		(* Do the work. *)
		if err = noErr then begin

			(* Use our resource file.  We need to do this because the client might go:
				Begin, switch resource chain order, GetPref, so we're not guaranteed
				to be at the top of this point.
			*)
			currentRefNum := CurResFile;
			UseResFile(globals^.configFileRefNum);
			err := ResError;
			if err = noErr then begin
				
				(* Get the existing preference and handle weird Resource Manager
					error conditions.
				*)
				prefResourceH := Get1NamedResource(globals^.configProfileResType, copiedKey);
				err := ResError;
				if prefResourceH = nil then begin
					err := icPrefNotFoundErr;
				end; (* if *)
				
				(* Get and check the preference resource for validity. It must contain
					at least 4 bytes of attributes at the beginning.
				*)
				if err = noErr then begin
					trueSizeOfPrefResource := GetHandleSize(prefResourceH);
					if trueSizeOfPrefResource < sizeof(ICAttr) then begin
						err := icPrefDataErr;
					end; (* if *)
				end; (* if *)
				
				(* Return the preference to the client. *)
				if err = noErr then begin
					size := trueSizeOfPrefResource - sizeof(ICAttr);
					attr := longintPtr(prefResourceH^)^;
					if (buf <> nil) and (size <> 0) then begin
						if size > maxSizeOfClientBuffer then begin
							err := icTruncatedErr;
						end else begin
							maxSizeOfClientBuffer := size;
						end; (* if *)
						BlockMoveData(Ptr(longint(prefResourceH^) + sizeof(ICAttr)), buf, maxSizeOfClientBuffer);
					end; (* if *)
				end; (* if *)
				UseResFile(currentRefNum);
			end; (* if *)
		end; (* if *)
		
		(* Clean up. *)
		if prefResourceH <> nil then begin
			ReleaseResource(prefResourceH);
		end; (* if *)
		err2 := ReleaseInside(globals, forceInsideInfo);
		if err = noErr then begin
			err := err2;
		end; (* if *)
		ICCIGetPref := err;
	end; (* ICCIGetPref *)

	function ICCISetPref(globals : ComponentGlobalsPtr; key: StringPtr; attr: ICAttr; buf: Ptr; size: longint): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			err2: ICError;
			junk : ICError;
			copiedKey: Str255;
			oldAttr: longint;
			currentRefNum: integer;
			prefH: Handle;
			forceInsideInfo: Boolean;
	begin
		(* Initialise our locals and return results to nice default values. *)
		copiedKey := '';
		prefH := nil;
		if buf = nil then begin
			(* API defines that if buf is nil, we're only setting the attributes
				and we should ignore size.  As we have a validity check for size
				later on, we want to make sure it'll pass that check.
			*)
			size := 0;
		end; (* if *)

		(* Make sure we're inside a read/write capable Begin/End pair. *)
		err := ForceInside(globals, icReadWritePerm, forceInsideInfo);
		if (err = noErr) and (globals^.configFilePerm <> icReadWritePerm) then begin
			err := icPermErr;
		end; (* if *)

		(* Check other parameters. *)
		if (err = noErr) and (key <> nil) then begin
			BlockMoveData(key, @copiedKey, ord(key^[0]) + 1);
		end; (* if *)
		if (err = noErr) and ((copiedKey = '') or (size < 0)) then begin
			(* key was nil, or key was empty, or size is bogus. *)
			err := paramErr;
		end; (* if *)
		
		(* Do the work. *)
		if err = noErr then begin

			(* Use our resource file.  We need to do this because the client might go:
				Begin, switch resource chain order, SetPref, so we're not guaranteed
				to be at the top of this point.
			*)
			currentRefNum := CurResFile;
			UseResFile(globals^.configFileRefNum);
			err := ResError;
			if err = noErr then begin

				(* Make sure the preference exists and is vaguely valid, ie
					has a size greater than 4.  If it doesn't, trash that resource
					and pretend we never saw it.
				*)
				prefH := Get1NamedResource(globals^.configProfileResType, copiedKey);
				if (prefH <> nil) & (GetHandleSize(prefH) < sizeof(ICAttr)) then begin { very bad! }
					RemoveResource(prefH);
					DisposeHandle(prefH);
					prefH := nil;
				end; (* if *)
				
				(* Work out the attributes that we're trying to set and check that
					this won't conflict with the permissions.
				*)
				if prefH = nil then begin
					oldAttr := 0;
				end else begin
					oldAttr := longintPtr(prefH^)^;
				end; (* if *)
				if attr = ICattr_no_change then begin
					attr := oldAttr;
				end; (* if *)
				if btst(oldAttr, ICattr_locked_bit) and
						 btst(attr, ICattr_locked_bit) and
						 		(buf <> nil) then begin
					err := icPermErr;
				end; (* if *)
				
				(* Note that we can't have an icPermErr if prefH is nil because
					prefH = nil => oldAttr = 0 => btst(oldAttr, ICattr_locked_bit) = false.
					So there's no possibility of the next chunk of code swallowing
					the error from the previous chunk of code.
				*)
				
				(* If we didn't find an old preference, create one and add it to
					the resource file, making sure we give it a unique ID.  Note
					that when we come out of this if statement, either
						o err = 0 and prefH is a valid resource handle, OR
						o err <> 0 and prefH is nil
				*)
				if prefH = nil then begin
					prefH := NewHandle(size + sizeof(ICAttr));
					err := MemError;
					if err = noErr then begin
						err := AddNamedResource(prefH, globals^.configProfileResType, copiedKey);
						if err <> noErr then begin
							DisposeHandle(prefH);
							prefH := nil;
						end; (* if *)
					end; (* if *)
				end; (* if *)

				(* Resize the preference handle to accomodate the new
					preference data.  Only do this if buf <> nil because the
					API says that buf=nil implies that we're only operating
					on the attributes not preferences data.
					Then copy the preference data into the new handle.
				*)
				if (err = noErr) & (buf <> nil) then begin
					SetHandleSize(prefH, size + sizeof(ICAttr));
					err := MemError;
				end; (* if *)
				if (err = noErr) & (size > 0) then begin
					BlockMoveData(buf, Ptr(longint(prefH^) + sizeof(ICAttr)), size);
				end; (* if *)
				
				(* If all goes well, copy in the new attributes and mark the resource
					as changed.
				*)
				if (err = noErr) then begin
					longintPtr(prefH^)^ := attr;
					ChangedResource(prefH);
					WriteResource(prefH);
					err := ResError;
				end; (* if *)
				UseResFile(currentRefNum);
			end; (* if *)
		end; (* if *)
		
		// We've set the preference correctly.  Now we run a tail check
		// to see whether we were settings the "Mappings" preference.
		// If so, and we has the default config file and there are
		// interrupt safe clients, let the interrupt
		// safe mapping code know.  This is a relatively evil piece of design,
		// but the alternative (a generic notification mechanism) is way too
		// hard at this stage.
		
		if err = noErr then begin
			if (buf <> nil) & 
						globals^.haveDefaultConfig & 
						(globals^.sysGlobals^.interruptSafeCount > 0) &
						EqualString(copiedKey, kICMapping, false, true) then begin
				junk := SetInterrruptSafeMapping(globals, buf, size);
				
				// If we fail, it's very likely because the system heap
				// is out of space.  We will just fail silently on non-debug
				// builds because it doesn't make sense to fail the entire
				// operation just because we can't tell interrupt safe clients
				// (which are few and far between) about it.  Chances are
				// the next update to the mappings (or the next reboot) will
				// cure this "out of sync" problems.
				//
				// On debug builds we'll stop at a watch point.
				
				if junk <> noErr then begin
					ComponentWatchPoint(globals, kICDebugSignificantEventsMask, 'Failed to update interrupt safe mappings', junk);
				end; (* if *)
			end; (* if *)
		end; (* if *)

		(* Clean up. *)
		if prefH <> nil then begin
			ReleaseResource(prefH);
		end; (* if *)
		err2 := ReleaseInside(globals, forceInsideInfo);
		if err = noErr then begin
			err := err2;
		end; (* if *)
		if err = noErr then begin
			DirtyThisFile(globals);
		end; (* if *)
		ICCISetPref := err;
	end; (* ICCISetPref *)

	(* ***** Reading and Writing Preference Handles ***** *)

	function ICCIGetPrefHandle (globals : ComponentGlobalsPtr; key: StringPtr; var attr: ICAttr; var prefh: Handle): ICError;
		(* See comment in ICAPI.p. This function is a deprecated form of FindPrefHandle.
			In this implementation, it provides a simple wrapper around FindPrefHandle
			that 'fixes' the behaviour.  Clients shouldn't really be calling this routine.
		*)
		var
			err: ICError;
	begin
		(* Allocates the handle in the current heap -- not a standard IC API practice. *)
		prefh := NewHandle(0);
		err := MemError;

		(* Call through to FindPrefHandle, which will do all the parameter validation. *)
		if err = noErr then begin
			err := ICCFindPrefHandle(globals^.target, key^, attr, prefh);
		end; (* if *)
		
		(* If the preferences isn't found, return noErr and an empty handle. *)
		if err = icPrefNotFoundErr then begin
			SetHandleSize(prefh, 0);
			attr := 0;
			err := noErr;
		end; (* if *)
		ICCIGetPrefHandle := err;
	end; (* ICCIGetPrefHandle *)

	function ICCIFindPrefHandle (globals : ComponentGlobalsPtr; key: StringPtr; var attr: ICAttr; prefh: Handle): ICError;
		(* See comment in ICAPI.p.  This routine calls ForceInside to speed things up.
			ForceInside will do an Begin and hence open the resource file, which is
			good because otherwise GetPref would open it up each time we call it,
			ie twice.
		*)
		var
			err: ICError;
			err2: ICError;
			preferenceSize: longint;
			forceInsideInfo: Boolean;
			s : SignedByte;
	begin
		(* Check the parameters. *)
		err := noErr;
		if prefh = nil then begin
			err := paramErr;
		end; (* if *)

		(* Do the work. *)
		preferenceSize := 0;
		if err = noErr then begin
			err := ForceInside(globals, icReadOnlyPerm, forceInsideInfo);
			s := HGetState(prefh);
			
			(* Get the size of the preference, resize the handle to that size,
				and then get the preference data into the handle.
			*)
			if err = noErr then begin
				err := ICCGetPref(globals^.target, key^, attr, nil, preferenceSize);
			end; (* if *)
			if err = noErr then begin
				HUnlock(prefh);
				SetHandleSize(prefh, preferenceSize);
				err := MemError;
			end; (* if *)
			if err = noErr then begin
				HLock(prefh);
				err := ICCGetPref(globals^.target, key^, attr, prefh^, preferenceSize);
			end; (* if *)

			HSetState(prefh, s);
			err2 := ReleaseInside(globals, forceInsideInfo);
			if err = noErr then begin
				err := err2;
			end; (* if *)
		end; (* if *)

		(* Clean up. *)
		if (prefh <> nil) and (err <> noErr) then begin
			SetHandleSize(prefh, 0);
		end; (* if *)
		ICCIFindPrefHandle := err;
	end; (* ICCIFindPrefHandle *)

	function ICCISetPrefHandle (globals : ComponentGlobalsPtr; key: StringPtr; attr: ICAttr; prefh: Handle): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			s: SignedByte;
	begin
		err := noErr;
		if prefh <> nil then begin
			if prefh^ = nil then begin
				err := paramErr;
			end; (* if *)
			if err = noErr then begin
				s := HGetState(prefh);
				HLock(prefh);
				err := ICCSetPref(globals^.target, key^, attr, prefh^, GetHandleSize(prefh));
				HSetState(prefh, s);
			end; (* if *)
		end else begin
			err := ICCSetPref(globals^.target, key^, attr, nil, 0);
		end; (* if *)
		ICCISetPrefHandle := err;
	end; (* ICCISetPrefHandle *)

	(* ***** Indexing Preferences ***** *)

	function ICCICountPref (globals : ComponentGlobalsPtr; var count: longint): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			currentRefNum: integer;
	begin
		err := CheckInside(globals, icReadOnlyPerm);
		if err = noErr then begin
			(* Switch the resource files then count the preference resources. *)
			currentRefNum := CurResFile;
			UseResFile(globals^.configFileRefNum);
			err := ResError;
			if err = noErr then begin
				count := Count1Resources(globals^.configProfileResType);
				err := ResError;
				UseResFile(currentRefNum);
			end; (* if *)
		end; (* if *)

		(* Clean up. *)
		if err <> noErr then begin
			count := 0;
		end; (* if *)
		ICCICountPref := err;
	end; (* ICCICountPref *)

	function ICCIGetIndPref (globals : ComponentGlobalsPtr; n: longint; var key: Str255): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			currentRefNum: integer;
			prefh: Handle;
			junkResourceID: integer;
			junkResourceType: ResType;
	begin
		(* Prepare for failure. *)
		prefh := nil;

		(* Check parameters. *)
		err := CheckInside(globals, icReadOnlyPerm);
		if (err = noErr) and (n < 1) then begin
			err := paramErr;						(* Bogus index parameter. *)
		end; (* if *)
		
		(* Do the work. *)
		if err = noErr then begin
			(* Switch resource files, get the resource, then ask for it's name. *)
			currentRefNum := CurResFile;
			UseResFile(globals^.configFileRefNum);
			err := ResError;
			if err = noErr then begin

				(* SetResLoad to false to avoid having the Resource Manager
					actually bring the preference data into RAM.  This reduces
					the footprint on the client application.
				*)
				SetResLoad(false);
				prefh := Get1IndResource(globals^.configProfileResType, n);
				SetResLoad(true);
				if prefh = nil then begin
					err := icPrefNotFoundErr;
				end else begin
					GetResInfo(prefh, junkResourceID, junkResourceType, key);
					err := ResError;
				end; (* if *)
				UseResFile(currentRefNum);
			end; (* if *)
		end; (* if *)

		(* Clean up. *)
		if prefh <> nil then begin
			ReleaseResource(prefh);
		end; (* if *)
		ICCIGetIndPref := err;
	end; (* ICCIGetIndPref *)

	(* ***** Deleting Preferences ***** *)

	function ICCIDeletePref (globals : ComponentGlobalsPtr; key: StringPtr): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			copiedKey: Str255;
			prefH: Handle;
			currentRefNum: integer;
	begin
		(* Check parameters. *)
		copiedKey := '';
		err := CheckInside(globals, icReadWritePerm);
		if (err = noErr) and (key <> nil) then begin
			BlockMoveData(key, @copiedKey, ord(key^[0]) + 1);
		end; (* if *)
		if (err = noErr) and (copiedKey = '') then begin
			(* key was nil, or key was empty. *)
			err := paramErr;
		end; (* if *)

		// We must prevent people from deleting all the preferences in
		// a profile.  The simplest way to do that is to dictate that you're
		// not allowed to delete the kICProfileName preference, which we
		// do right here.
		
		if (err = noErr) & EqualString(copiedKey, kICProfileName, false, true) then begin
			err := icPermErr;
		end; (* if *)
		
		(* Do the work. *)
		if err = noErr then begin
			(* Switch resource files, get the resource, then remove it. *)
			currentRefNum := CurResFile;
			UseResFile(globals^.configFileRefNum);
			err := ResError;
			if err = noErr then begin
				
				(* SetResLoad to false to avoid having the Resource Manager
					actually bring the preference data into RAM.  This reduces
					the footprint on the client application.
				*)
				SetResLoad(false);
				prefH := Get1NamedResource(globals^.configProfileResType, copiedKey);
				err := ResError;
				SetResLoad(true);
				if (prefH = nil) or (err = resNotFound) then begin
					(* If the preference is missing, we want to return the
						correct IC error code.
					*)
					err := icPrefNotFoundErr;
				end; (* if *)

				(* Zap the preference. *)
				if err = noErr then begin
					RemoveResource(prefH);
					err := ResError;
				end; (* if *)
				UseResFile(currentRefNum);
			end; (* if *)
		end; (* if *)
		ICCIDeletePref := err;
	end; (* ICCIDeletePref *)

end. (* ICCore *)